
//*****************************************************************************************//
//                                      Includes                                           //
//*****************************************************************************************//

#include "includes.all"

//*****************************************************************************************//
//                                    FrameBuffers                                         //
//*****************************************************************************************//

FrameBuffer::FrameBuffer (long width, long height, bool isCubeMap) {
	this->width = width; this->height = height; this->isCubeMap = isCubeMap;
	textureType = isCubeMap ? GL_TEXTURE_CUBE_MAP : GL_TEXTURE_2D;
	isActivated = false; hasColorAttachment = hasDepthAttachment = false;
	frameBufferHandle = 0; isBroken = false;
}

void FrameBuffer::setFiltering (GLint filter, GLint clamping, bool mipmap) {
    glTexParameteri (textureType, GL_TEXTURE_MIN_FILTER, filter);
    glTexParameteri (textureType, GL_TEXTURE_MAG_FILTER, filter);
	glTexParameteri (textureType, GL_TEXTURE_WRAP_S, clamping);
    glTexParameteri (textureType, GL_TEXTURE_WRAP_T, clamping);
	glTexParameteri (textureType, GL_GENERATE_MIPMAP, mipmap);
}

void FrameBuffer::attachDepthOnly () {
	detach (); 

	glGenTextures (1, &depthTextureAttachmentHandle);
	glBindTexture (textureType, depthTextureAttachmentHandle);

	//Create a special texture for depth
	glTexImage2D (textureType, 0, GL_DEPTH_COMPONENT, width, height, 0, GL_DEPTH_COMPONENT, GL_FLOAT, 0);
	setFiltering (GL_LINEAR, GL_CLAMP, false);
	glBindTexture (textureType, 0);

	//Generating ID
	glGenFramebuffersEXT (1, &frameBufferHandle);
	activate ();

	//We attach the texture to the depth attachment point
	glFramebufferTexture2DEXT (GL_FRAMEBUFFER_EXT, GL_DEPTH_ATTACHMENT_EXT, textureType, depthTextureAttachmentHandle, 0);

	//In order to avoid Color in the depth buffer
	glDrawBuffer (GL_NONE); glReadBuffer (GL_NONE);

	deactivate ();

	//Check frame buffer status.
	isBroken = !isComplete ();
	
	//Depth Only
	hasColorAttachment = false; hasDepthAttachment = true;
}

void FrameBuffer::attachColorOnly (bool use16Bits) {
    detach (); 

	//Generate the texture
	glGenTextures (1, &colorTextureAttachmentHandle);
	glBindTexture (textureType, colorTextureAttachmentHandle);

	setFiltering (GL_LINEAR, GL_CLAMP_TO_EDGE, true);

	///We need six Face for cube, or one for 2D
	if (textureType == GL_TEXTURE_CUBE_MAP) {
		glTexImage2D (GL_TEXTURE_CUBE_MAP_POSITIVE_X , 0, GL_RGBA8, width, height, 0, GL_RGBA, GL_UNSIGNED_BYTE, 0);
		glTexImage2D (GL_TEXTURE_CUBE_MAP_POSITIVE_Y , 0, GL_RGBA8, width, height, 0, GL_RGBA, GL_UNSIGNED_BYTE, 0);
		glTexImage2D (GL_TEXTURE_CUBE_MAP_POSITIVE_Z , 0, GL_RGBA8, width, height, 0, GL_RGBA, GL_UNSIGNED_BYTE, 0);
		glTexImage2D (GL_TEXTURE_CUBE_MAP_NEGATIVE_X , 0, GL_RGBA8, width, height, 0, GL_RGBA, GL_UNSIGNED_BYTE, 0);
		glTexImage2D (GL_TEXTURE_CUBE_MAP_NEGATIVE_Y , 0, GL_RGBA8, width, height, 0, GL_RGBA, GL_UNSIGNED_BYTE, 0);
		glTexImage2D (GL_TEXTURE_CUBE_MAP_NEGATIVE_Z , 0, GL_RGBA8, width, height, 0, GL_RGBA, GL_UNSIGNED_BYTE, 0);
	} else {
		glTexImage2D (textureType, 0, use16Bits ? GL_RGB16F_ARB : GL_RGBA8, width, height, 0, GL_RGBA, GL_UNSIGNED_BYTE, 0);
	}
	
	glBindTexture (textureType, 0);

	//Generate renderbuffer
	glGenRenderbuffersEXT (1, &depthBufferAttachmentHandle);
    glBindRenderbufferEXT (GL_RENDERBUFFER_EXT, depthBufferAttachmentHandle);
    glRenderbufferStorageEXT (GL_RENDERBUFFER_EXT, GL_DEPTH_COMPONENT24, width, height);
    glBindRenderbufferEXT (GL_RENDERBUFFER_EXT, 0);

	//Generating ID
	glGenFramebuffersEXT (1, &frameBufferHandle);
	activate ();

    //Attach a texture to frame buffer color attachement point
    glFramebufferTexture2DEXT (GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, textureType, colorTextureAttachmentHandle, 0);

	//Attach a renderbuffer to depth attachment point
    glFramebufferRenderbufferEXT (GL_FRAMEBUFFER_EXT, GL_DEPTH_ATTACHMENT_EXT, GL_RENDERBUFFER_EXT, depthBufferAttachmentHandle);

	deactivate ();

	//Check frame buffer status.
	isBroken = !isComplete ();
	
	//Color Only
	hasColorAttachment = true; hasDepthAttachment = false;
}

void FrameBuffer::attachColorAndDepth () {
    detach ();

	//Create a texture object for the depthmap
    glGenTextures (1, &depthTextureAttachmentHandle);
	glBindTexture (textureType, depthTextureAttachmentHandle);
	setFiltering (GL_LINEAR, GL_CLAMP, true);
    glTexImage2D (textureType, 0, GL_DEPTH_COMPONENT, width, height, 0, GL_DEPTH_COMPONENT, GL_UNSIGNED_BYTE, 0);
    glBindTexture (textureType, 0);

    //create a texture object
    glGenTextures (1, &colorTextureAttachmentHandle);
    glBindTexture (textureType, colorTextureAttachmentHandle);
	setFiltering (GL_LINEAR, GL_CLAMP_TO_EDGE, true);
    glTexImage2D (textureType, 0, GL_RGBA8, width, height, 0, GL_RGBA, GL_UNSIGNED_BYTE, 0);
    glBindTexture (textureType, 0);

    glGenFramebuffersEXT (1, &frameBufferHandle);
    activate ();

    //Attach a texture to frame buffer color attachement point
    glFramebufferTexture2DEXT (GL_FRAMEBUFFER_EXT, GL_DEPTH_ATTACHMENT_EXT, textureType, depthTextureAttachmentHandle, 0);
    glFramebufferTexture2DEXT (GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, textureType, colorTextureAttachmentHandle, 0);

	deactivate ();

	//Check frame buffer status.
	isBroken = !isComplete ();
	
	//Color + Depth
	hasColorAttachment = hasDepthAttachment = true;
}

void FrameBuffer::detach () {
	if (frameBufferHandle == 0) return;
	glDeleteFramebuffersEXT (1, &frameBufferHandle);

	if (hasColorAttachment)
		glDeleteTextures (1, &colorTextureAttachmentHandle);

	if (hasDepthAttachment)
		glDeleteRenderbuffersEXT (1, &depthTextureAttachmentHandle);
	else
		glDeleteRenderbuffersEXT (1, &depthBufferAttachmentHandle);
	
	isActivated = false;
}

void FrameBuffer::activateAsTexture () {
	if (isBroken) {glDisable (textureType); return;}
	if (hasColorAttachment) {
		glEnable (textureType);
		glBindTexture (textureType, colorTextureAttachmentHandle);
	}
}
void FrameBuffer::deactivateAsTexture () {
	glBindTexture (textureType, 0);
	glDisable (textureType);
}

void FrameBuffer::activateAsDepthTexture () {
	if (isBroken) {glDisable (textureType); return;}
	if (hasDepthAttachment) {
		glEnable (textureType);
		glBindTexture (textureType, depthTextureAttachmentHandle);
	}
}
void FrameBuffer::deactivateAsDepthTexture () {
	deactivateAsTexture ();
}

void FrameBuffer::activate () {
	if (isBroken) {deactivate (); return;}
	isActivated = true;
	glBindFramebufferEXT (GL_FRAMEBUFFER_EXT, frameBufferHandle);
}

void FrameBuffer::activate (long faceIndex) {//0,1,2,3,4, or 5.
	if (isBroken) {glDisable (textureType); return;}
	activate ();
	if (!isCubeMap) return;
	glFramebufferTexture2DEXT (GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, GL_TEXTURE_CUBE_MAP_POSITIVE_X+faceIndex, colorTextureAttachmentHandle, 0);
}

void FrameBuffer::deactivate () {
	isActivated = false;
	glBindFramebufferEXT (GL_FRAMEBUFFER_EXT, 0);
}

bool FrameBuffer::isComplete () {
    GLenum status = glCheckFramebufferStatusEXT (GL_FRAMEBUFFER_EXT);
	char *message;
    switch (status) {
		case GL_FRAMEBUFFER_COMPLETE_EXT: 
			return true; ;
		case GL_FRAMEBUFFER_INCOMPLETE_ATTACHMENT_EXT:
			message = "INCOMPLETE_ATTACHMENT"; break;
		case GL_FRAMEBUFFER_UNSUPPORTED_EXT:
			message = "UNSUPPORTED"; break;
		case GL_FRAMEBUFFER_INCOMPLETE_MISSING_ATTACHMENT_EXT:
			message = "INCOMPLETE_MISSING_ATTACHMENT"; break;
		case GL_FRAMEBUFFER_INCOMPLETE_DIMENSIONS_EXT:
			message = "INCOMPLETE_DIMENSIONS"; break;
		case GL_FRAMEBUFFER_INCOMPLETE_FORMATS_EXT:
			message = "INCOMPLETE_FORMATS"; break;
		case GL_FRAMEBUFFER_INCOMPLETE_DRAW_BUFFER_EXT:
			message = "INCOMPLETE_DRAW_BUFFER"; break;
		case GL_FRAMEBUFFER_INCOMPLETE_READ_BUFFER_EXT:
			message = "INCOMPLETE_READ_BUFFER"; break;
		default:
			::log ("\nIncomplete frame buffer with unknown error \"%x\"...", status);
			return false;
    }
	::log ("\nIncomplete frame buffer with error \"%s\"...", message);
	return false;
}

void FrameBuffer::example () {
	//This example is conceptual; i.e., you can't actually run it...
	return; //Just in case you try...
	
	//Do the following in a setup routine so it's only ever done ONCE...
	FrameBuffer *frameBuffer = new FrameBuffer (1024, 768);
	frameBuffer->attachColorOnly ();

	//Do the following each time you want to draw into the frame buffer.
	frameBuffer->activate ();
		glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
		//Setup all your transformations (including projection) and draw anything you want now...
	frameBuffer->deactivate ();

	//Do the following each time you want to use the frame buffer as if it were a texture.
	glActiveTexture (GL_TEXTURE0);
	frameBuffer->activateAsTexture ();
		//Draw something that does not itself activate it's own texture (it will use this one instead).
		//You could use additional textures via 	glActiveTexture (GL_TEXTURE1); //Then 2, then 3, ...
	frameBuffer->deactivateAsTexture ();
	
	//Do the following in a wrapup routine so it's only ever done ONCE...
	delete frameBuffer;
}