//*****************************************************************************************//
//                                     Utilities                                           //
//*****************************************************************************************//

#ifndef utilitiesModule
#define utilitiesModule 

//Macro "declareCollection" below declares a type of vector, e.g., declareCollection (Face);
//creates a vector of Face pointers.

//There are two ways to use such a collection

//	1. Typical way of using a collection:
//
//			FaceCollection faces; 
//			Face *face1 = new Face; ...; faces.push_back (face1); 
//			Face *face2 = new Face; ...; faces.push_back (face2);
//				...
//			loopVector (index, faces)
//				... faces [index] .... //provides an element of type Face *.
//			endloop
//			deleteFaceCollectionEntries (faces); //faces is deleted when the object containing it is deleted...

//	2. More unusual way (using a pointer to a collection):
//
//			FaceCollection *faces = new FaceCollection; 
//			Face *face1 = new Face; ...; faces->push_back (face1); 
//			Face *face2 = new Face; ...; faces->push_back (face2);
//				...
//			loopVector (index, *faces)
//				... (*faces) [index] .... //provides an element of type Face *.
//			endloop
//			deleteFaceCollection (faces); //deletes the entries and also faces itself...

#define declareCollection(type) \
	typedef vector <type *> type##Collection; \
	typedef vector <type *>::iterator type##CollectionIterator; \
		\
	inline void delete##type##CollectionEntries (type##Collection &collection) { \
		for (long index = 0; index < collection.size (); index++) { \
			delete collection [index]; \
		} \
	} \
	inline void delete##type##Collection (type##Collection *collection) { \
		delete##type##CollectionEntries (*collection); \
		delete collection; \
	}

#define loopVector(variable,vector) {for (long variable = 0; variable < (vector).size (); variable++) {
#define endloop }}



//Create a special kind of StringCollection for dealing with string values...
//It eliminates the need to allocate or delete the values since they are 
//handled automatically by the collection.

typedef vector <string> StringCollection; 
typedef vector <string>::iterator StringCollectionIterator; 
	
inline void deleteStringCollection (StringCollection *collection) { 
	delete collection; 
}

//Example use:
//	StringCollection test;
//	test.push_back ("chair");
//	test.push_back ("brick");
//	test.push_back ("house");
//	loopVector (index, test)
//		if (stricmp (test [index], "brick") == 0) ::log ("\nBrick is at index %d.", index);
//		::log ("\nString %d is \"%s\"...", index, test [index]);
//	endloop





//Macro "declareDictionary" below declares a type of map, e.g., declareDictionary (Face)
//creates a map that associates strings with Face pointers.

//There are two ways to use such a collection

//	1. Typical way of using a dictionary:
//
//			FaceDictionary faces; 
//			Face *face1 = new Face; ...; faces ["name1"] = face1; //Storing
//			Face *face2 = new Face; ...; faces ["name2"] = face2; 
//				...
//			Face *face = faces ["name1"]; //Retrieving: get NULL if not found (also creates key "name1")...
//			if (key_exists (faces, "name1") ... //Safer (no new keys created).

//			faces ["name1"] = face2; //Changing... but will cause memory leak unless "delete faces ["name1"]" done first.
//			loopDictionary (faceKey, faceValue, faces, Face)
//				... faceValue .... //faceValue is locally declared by the loop macro to be type Face * and initialized...
//			endloop
//			deleteFaceDictionaryEntries (faces); //faces is deleted when the object containing it is deleted...

//	2. More unusual way (using a pointer to a collection):
//
//			FaceDictionary *faces = new FaceDictionary; 
//			As above but use "faces->" instead of "faces." or *faces instead of faces.
//			deleteFaceDictionary (faces); //deletes the entries and also faces itself...

#define loopDictionary(key,value,map,type) \
	{for (type##DictionaryIterator iterator = (map).begin (); iterator != (map).end (); ++iterator) { \
		const char *key = iterator->first.c_str (); \
		type *value = iterator->second;

#define declareDictionary(type) \
	typedef map <string, type *> type##Dictionary; \
	typedef type##Dictionary::iterator type##DictionaryIterator; \
		\
	inline void delete##type##DictionaryEntries (type##Dictionary &collection) { \
		loopDictionary (key, value, collection, type) \
			delete value; \
		endloop \
	} \
	inline void delete##type##Dictionary (type##Dictionary *collection) { \
		delete##type##DictionaryEntries (*collection); \
		delete collection; \
	} \
		\
	inline bool keyExists (type##Dictionary &collection, char *key) { \
		type##DictionaryIterator result = collection.find (key); \
		return result != collection.end (); \
	} 

//Create a special kind of StringDictionary for dealing with properties; i.e. 
//dictionaries that have string keys and string values...

//It eliminates the need to allocate or delete the values since they are 
//handled automatically by the collection (just like the keys are for the 
//standard types of dictionaries).

#define loopStringDictionary(key,value,map) \
	{for (StringDictionaryIterator iterator = (map).begin (); iterator != (map).end (); ++iterator) { \
		const char *key = iterator->first.c_str (); \
		const char *value = iterator->second.c_str ();


typedef map <string, string> StringDictionary; 
typedef StringDictionary::iterator StringDictionaryIterator; 
	
inline bool keyExists (StringDictionary &collection, char *key) { 
	StringDictionaryIterator result = collection.find (key); 
	return result != collection.end (); 
} 

//Example use:
//	StringDictionary test;
//	test ["a"] = "chair");
//	test ["b"] = "brick");
//	test ["c"] = "house");
//	loopStringDictionary (key, value, test)
//		if (stricmp (value, "brick") == 0) ::log ("\nBrick is at key %s.", key);
//		::log ("\nKey %s has vlue %s...", key, value);
//	endloop
//	const char *test1 = test ["a"].c_str (); //Retrieves "chair".
//	const char *test2 = test ["x"].c_str (); //Careful: Missing key creates a new key with value "".
//	if (keyExists (test, "y)) ... //Returns false without creating a new key...


//If you need a forward reference to a collection, you will need the following macros.
//An example would be if class Object itself needs to keep track of a collection of sub-objects.
//In that case, you would need to declare a collection of objects but you will have difficulty
//doing this before the class definition...

#define preDeclareCollection(classOrStruct,type) \
	classOrStruct type; \
	typedef vector <type *> type##Collection;

#define preDeclareDictionary(classOrStruct,type) \
	classOrStruct type; \
	typedef map <string, type *> type##Dictionary;


//Generic macros for reading universal files...
//Note: input.getline (line, 256, c) => picks up everything up to c exclusive and discards c.

#define SKIP_TO(character) input.getline (line, 256, character)
#define SKIP_TO_ENDLINE SKIP_TO ('\n')
#define SKIP_TO_COMMA SKIP_TO (',')
#define SKIP_TO_COLON SKIP_TO (':')
#define SKIP_TO_SEMICOLON SKIP_TO (';')
#define CLEAR_THE_LINE SKIP_TO_ENDLINE

//Useful utility...
void convertToLowercase (char *name);
bool isExtensionSupported (const char *allExtensions, const char *extension);
void logExtensions (bool logAllExtensions);

inline long minimum (long a, long b) {return a <= b ? a : b;}
inline double minimum (double a, double b) {return a <= b ? a : b;}
inline long maximum (long a, long b) {return a >= b ? a : b;}
inline double maximum (double a, double b) {return a >= b ? a : b;}

//ARB vertex buffer objects
//#include "ARBvertexBufferObject.h"
extern PFNGLBINDBUFFERARBPROC glBindBufferARB;
extern PFNGLDELETEBUFFERSARBPROC glDeleteBuffersARB;
extern PFNGLGENBUFFERSARBPROC glGenBuffersARB;
extern PFNGLISBUFFERARBPROC glIsBufferARB;
extern PFNGLBUFFERDATAARBPROC glBufferDataARB;
extern PFNGLBUFFERSUBDATAARBPROC glBufferSubDataARB;
extern PFNGLGETBUFFERSUBDATAARBPROC glGetBufferSubDataARB;
extern PFNGLMAPBUFFERARBPROC glMapBufferARB;
extern PFNGLUNMAPBUFFERARBPROC glUnmapBufferARB;
extern PFNGLGETBUFFERPARAMETERIVARBPROC glGetBufferParameterivARB;
extern PFNGLGETBUFFERPOINTERVARBPROC glGetBufferPointervARB;

extern PFNGLATTACHOBJECTARBPROC glAttachObjectARB;
extern PFNGLCOMPILESHADERARBPROC glCompileShaderARB;
extern PFNGLCREATEPROGRAMOBJECTARBPROC glCreateProgramObjectARB;
extern PFNGLCREATESHADEROBJECTARBPROC glCreateShaderObjectARB;
extern PFNGLDELETEOBJECTARBPROC glDeleteObjectARB;
extern PFNGLDETACHOBJECTARBPROC glDetachObjectARB;
extern PFNGLGETACTIVEUNIFORMARBPROC glActiveUniformARB;
extern PFNGLGETATTACHEDOBJECTSARBPROC glGetAttachedObjectARB;
extern PFNGLGETHANDLEARBPROC glGetHandleARB;
extern PFNGLGETINFOLOGARBPROC glGetInfoLogARB;
extern PFNGLGETOBJECTPARAMETERFVARBPROC glGetObjectParameterfvARB;
extern PFNGLGETOBJECTPARAMETERIVARBPROC glGetObjectParameterivARB;
extern PFNGLGETSHADERSOURCEARBPROC glGetShaderSourceARB;
extern PFNGLGETUNIFORMLOCATIONARBPROC glGetUniformLocationARB;
extern PFNGLGETUNIFORMFVARBPROC glGetUniformfvARB;
extern PFNGLGETUNIFORMIVARBPROC glGetUniformivARB;
extern PFNGLLINKPROGRAMARBPROC glLinkProgramARB;
extern PFNGLSHADERSOURCEARBPROC glShaderSourceARB;
extern PFNGLUNIFORM1FARBPROC glUniform1fARB;
extern PFNGLUNIFORM1FVARBPROC glUniform1fvARB;
extern PFNGLUNIFORM1IARBPROC glUniform1iARB;
extern PFNGLUNIFORM1IVARBPROC glUniform1ivARB;
extern PFNGLUNIFORM2FARBPROC glUniform2fARB;
extern PFNGLUNIFORM2FVARBPROC glUniform2fvARB;
extern PFNGLUNIFORM2IARBPROC glUniform2iARB;
extern PFNGLUNIFORM2IVARBPROC glUniform2ivARB;
extern PFNGLUNIFORM3FARBPROC glUniform3fARB;
extern PFNGLUNIFORM3FVARBPROC glUniform3fvARB;
extern PFNGLUNIFORM3IARBPROC glUniform3iARB;
extern PFNGLUNIFORM3IVARBPROC glUniform3ivARB;
extern PFNGLUNIFORM4FARBPROC glUniform4fARB;
extern PFNGLUNIFORM4FVARBPROC glUniform4fvARB;
extern PFNGLUNIFORM4IARBPROC glUniform4iARB;
extern PFNGLUNIFORM4IVARBPROC glUniform4ivARB;
extern PFNGLUNIFORMMATRIX2FVARBPROC glUniformMatrix2ARB;
extern PFNGLUNIFORMMATRIX3FVARBPROC glUniformMatrix3ARB;
extern PFNGLUNIFORMMATRIX4FVARBPROC glUniformMatrix4ARB;
extern PFNGLUSEPROGRAMOBJECTARBPROC glUseProgramObjectARB;
extern PFNGLVALIDATEPROGRAMARBPROC glValidateProgramARB;

void setupOpenOLDStyleOpenGLExtensions (); //Routine to make make broken openGL routines work...

#endif //utilitiesModule